@@ -39,6 +39,7 @@ main (int argc, char **argv)
 #include <errno.h>
 #include <fcntl.h>
 #include <libunwind-ptrace.h>
+#include <signal.h>
 #include <stdio.h>
 #include <stdlib.h>
 #include <string.h>
@@ -47,6 +48,10 @@ main (int argc, char **argv)
 #include <sys/ptrace.h>
 #include <sys/wait.h>
 
+extern char **environ;
+
+static const int nerrors_max = 100;
+
 int nerrors;
 int verbose;
 int print_names = 1;
@@ -59,20 +64,23 @@ enum
   }
 trace_mode = SYSCALL;
 
-#define panic(args...)					\
+#define panic(args...)						\
 	do { fprintf (stderr, args); ++nerrors; } while (0)
 
 static unw_addr_space_t as;
 static struct UPT_info *ui;
 
+static int killed;
+
 void
 do_backtrace (pid_t target_pid)
 {
+  unw_word_t ip, sp, start_ip = 0, off;
   int n = 0, ret;
   unw_proc_info_t pi;
-  unw_word_t ip, sp;
   unw_cursor_t c;
   char buf[512];
+  size_t len;
 
   ret = unw_init_remote (&c, as, ui);
   if (ret < 0)
@@ -84,15 +92,27 @@ do_backtrace (pid_t target_pid)
 	  || (ret = unw_get_reg (&c, UNW_REG_SP, &sp)) < 0)
 	panic ("unw_get_reg/unw_get_proc_name() failed: ret=%d\n", ret);
 
+      if (n == 0)
+	start_ip = ip;
+
       buf[0] = '\0';
       if (print_names)
-	unw_get_proc_name (&c, buf, sizeof (buf), NULL);
+	unw_get_proc_name (&c, buf, sizeof (buf), &off);
 
       if (verbose)
-	printf ("%016lx %-32s (sp=%016lx)\n", (long) ip, buf, (long) sp);
+	{
+	  if (off)
+	    {
+	      len = strlen (buf);
+	      if (len >= sizeof (buf) - 32)
+		len = sizeof (buf) - 32;
+	      sprintf (buf + len, "+0x%lx", (unsigned long) off);
+	    }
+	  printf ("%016lx %-32s (sp=%016lx)\n", (long) ip, buf, (long) sp);
+	}
 
       if ((ret = unw_get_proc_info (&c, &pi)) < 0)
-	panic ("unw_get_proc_info() failed: ret=%d\n", ret);
+	panic ("unw_get_proc_info(ip=0x%lx) failed: ret=%d\n", (long) ip, ret);
       else if (verbose)
 	printf ("\tproc=%016lx-%016lx\n\thandler=%lx lsda=%lx",
 		(long) pi.start_ip, (long) pi.end_ip,
@@ -115,14 +135,20 @@ do_backtrace (pid_t target_pid)
       if (ret < 0)
 	{
 	  unw_get_reg (&c, UNW_REG_IP, &ip);
-	  panic ("FAILURE: unw_step() returned %d for ip=%lx\n",
-		 ret, (long) ip);
+	  panic ("FAILURE: unw_step() returned %d for ip=%lx (start ip=%lx)\n",
+		 ret, (long) ip, (long) start_ip);
 	}
 
-      if (++n > 32)
+      if (++n > 64)
 	{
 	  /* guard against bad unwind info in old libraries... */
-	  panic ("too deeply nested---assuming bogus unwind\n");
+	  panic ("too deeply nested---assuming bogus unwind (start ip=%lx)\n",
+		 (long) start_ip);
+	  break;
+	}
+      if (nerrors > nerrors_max)
+        {
+	  panic ("Too many errors (%d)!\n", nerrors);
 	  break;
 	}
     }
@@ -135,11 +161,16 @@ do_backtrace (pid_t target_pid)
     printf ("================\n\n");
 }
 
+static pid_t target_pid;
+static void target_pid_kill (void)
+{
+  kill (target_pid, SIGKILL);
+}
+
 int
 main (int argc, char **argv)
 {
   int status, pid, pending_sig, optind = 1, state = 1;
-  pid_t target_pid;
 
   as = unw_create_addr_space (&_UPT_accessors, 0);
   if (!as)
@@ -162,7 +193,8 @@ main (int argc, char **argv)
 	else if (strcmp (argv[optind], "-s") == 0)
 	  ++optind, trace_mode = SYSCALL;	/* backtrace at each syscall */
 	else if (strcmp (argv[optind], "-t") == 0)
-	  /* Execute until syscall(-1), then backtrace at each insn.  */
+	  /* Execute until raise(SIGUSR1), then backtrace at each insn
+	     until raise(SIGUSR2).  */
 	  ++optind, trace_mode = TRIGGER;
 	else if (strcmp (argv[optind], "-c") == 0)
 	  /* Enable caching of unwind-info.  */
@@ -170,6 +202,8 @@ main (int argc, char **argv)
 	else if (strcmp (argv[optind], "-n") == 0)
 	  /* Don't look-up and print symbol names.  */
 	  ++optind, print_names = 0;
+	else
+	  fprintf(stderr, "unrecognized option: %s\n", argv[optind++]);
       }
 
   target_pid = fork ();
@@ -180,14 +214,21 @@ main (int argc, char **argv)
       if (!verbose)
 	dup2 (open ("/dev/null", O_WRONLY), 1);
 
+#if HAVE_DECL_PTRACE_TRACEME
       ptrace (PTRACE_TRACEME, 0, 0, 0);
+#elif HAVE_DECL_PT_TRACE_ME
+      ptrace (PT_TRACE_ME, 0, 0, 0);
+#else
+#error Trace me
+#endif
       execve (argv[optind], argv + optind, environ);
       _exit (-1);
     }
+  atexit (target_pid_kill);
 
   ui = _UPT_create (target_pid);
 
-  while (1)
+  while (nerrors <= nerrors_max)
     {
       pid = wait4 (-1, &status,  0, 0);
       if (pid == -1)
@@ -209,12 +250,16 @@ main (int argc, char **argv)
 	    }
 	  else if (WIFSIGNALED (status))
 	    {
-	      panic ("child terminated by signal %d\n", WTERMSIG (status));
+	      if (!killed)
+		panic ("child terminated by signal %d\n", WTERMSIG (status));
 	      break;
 	    }
 	  else
 	    {
 	      pending_sig = WSTOPSIG (status);
+	      /* Avoid deadlock:  */
+	      if (WSTOPSIG (status) == SIGKILL)
+	        break;
 	      if (trace_mode == TRIGGER)
 		{
 		  if (WSTOPSIG (status) == SIGUSR1)
@@ -222,6 +267,17 @@ main (int argc, char **argv)
 		  else if  (WSTOPSIG (status) == SIGUSR2)
 		    state = 1;
 		}
+	      if (WSTOPSIG (status) != SIGUSR1 && WSTOPSIG (status) != SIGUSR2)
+	        {
+		  static int count = 0;
+
+		  if (count++ > 100)
+		    {
+		      panic ("Too many child unexpected signals (now %d)\n",
+			     WSTOPSIG (status));
+			killed = 1;
+		    }
+	        }
 	    }
 	}
 
@@ -229,11 +285,23 @@ main (int argc, char **argv)
 	{
 	case TRIGGER:
 	  if (state)
+#if HAVE_DECL_PTRACE_CONT
 	    ptrace (PTRACE_CONT, target_pid, 0, 0);
+#elif HAVE_DECL_PT_CONTINUE
+	    ptrace (PT_CONTINUE, target_pid, (caddr_t)1, 0);
+#else
+#error Port me
+#endif
 	  else
 	    {
 	      do_backtrace (target_pid);
+#if HAVE_DECL_PTRACE_SINGLESTEP
 	      ptrace (PTRACE_SINGLESTEP, target_pid, 0, pending_sig);
+#elif HAVE_DECL_PT_STEP
+	      ptrace (PT_STEP, target_pid, (caddr_t)1, pending_sig);
+#else
+#error Singlestep me
+#endif
 	    }
 	  break;
 
@@ -241,14 +309,28 @@ main (int argc, char **argv)
 	  if (!state)
 	    do_backtrace (target_pid);
 	  state ^= 1;
+#if HAVE_DECL_PTRACE_SYSCALL
 	  ptrace (PTRACE_SYSCALL, target_pid, 0, pending_sig);
+#elif HAVE_DECL_PT_SYSCALL
+	  ptrace (PT_SYSCALL, target_pid, (caddr_t)1, pending_sig);
+#else
+#error Syscall me
+#endif
 	  break;
 
 	case INSTRUCTION:
 	  do_backtrace (target_pid);
-	  ptrace (PTRACE_SINGLESTEP, target_pid, 0, pending_sig);
+#if HAVE_DECL_PTRACE_SINGLESTEP
+	      ptrace (PTRACE_SINGLESTEP, target_pid, 0, pending_sig);
+#elif HAVE_DECL_PT_STEP
+	      ptrace (PT_STEP, target_pid, (caddr_t)1, pending_sig);
+#else
+#error Singlestep me
+#endif
 	  break;
 	}
+      if (killed)
+        kill (target_pid, SIGKILL);
     }
 
   _UPT_destroy (ui);
